class ApiException extends Error {
    // @ts-ignore
    key;
    constructor(message) {
        super(message);
        this.message = message;
    }
}
class ValidationError extends ApiException {
    constructor(message) {
        super(message);
    }
}

class BaseParameter {
    static isParameter = true;
    isParameter = true;
    type = 'string';
    params;
    generated;
    constructor(params) {
        this.params = params || {};
        this.generated = true;
        if (this.params.required === undefined)
            this.params.required = true;
    }
    getValue() {
        const value = {
            type: this.type,
            description: this.params.description,
            example: this.params.example,
            default: this.params.default,
        };
        if (this.params.deprecated)
            value.deprecated = this.params.deprecated;
        return value;
    }
    validate(value) {
        return value;
    }
}
class Arr extends BaseParameter {
    isArr = true;
    innerType;
    constructor(innerType, params) {
        super(params);
        this.innerType = innerType;
    }
    validate(value) {
        value = super.validate(value);
        if (this.params.required === false && (value === null || value === '')) {
            return null;
        }
        if (Array.isArray(value)) {
            value = value.map((val) => {
                return this.innerType.validate(val);
            });
        }
        else {
            value = [this.innerType.validate(value)];
        }
        return value;
    }
    // @ts-ignore
    getValue() {
        return {
            type: 'array',
            items: this.innerType.getValue(),
        };
    }
}
class Obj extends BaseParameter {
    isObj = true;
    fields;
    constructor(fields, params) {
        super(params); // TODO: fix obj params
        this.fields = fields;
    }
    validate(value) {
        value = super.validate(value);
        for (const [key, param] of Object.entries(this.fields)) {
            try {
                if (value[key] === undefined || value[key] === null) {
                    if (param.params.required) {
                        throw new ValidationError('is required');
                    }
                }
                else {
                    value[key] = param.validate(value[key]);
                }
            }
            catch (e) {
                // @ts-ignore
                e.key = (e.key || '') + `.${key}`;
                throw e;
            }
        }
        return value;
    }
    // @ts-ignore
    getValue() {
        const result = {
            type: 'object',
            properties: {},
        };
        const required = [];
        for (const [key, value] of Object.entries(this.fields)) {
            if (value.params?.required === true) {
                required.push(key);
            }
            if (value.getValue) {
                result.properties[key] = value.getValue();
            }
            else {
                result.properties[key] = value;
            }
        }
        if (required.length > 0) {
            result.required = required;
        }
        if (this.params.xml) {
            result.xml = this.params.xml;
        }
        return result;
    }
}
class Num extends BaseParameter {
    type = 'number';
    validate(value) {
        value = super.validate(value);
        value = Number.parseFloat(value);
        if (isNaN(value)) {
            throw new ValidationError('is not a valid number');
        }
        return value;
    }
}
class Int extends Num {
    type = 'integer';
    validate(value) {
        value = super.validate(value);
        value = Number.parseInt(value);
        if (isNaN(value)) {
            throw new ValidationError('is not a valid integer');
        }
        return value;
    }
}
class Str extends BaseParameter {
    type = 'string';
    constructor(params) {
        super(params);
    }
    validate(value) {
        value = super.validate(value);
        if (typeof value !== 'string') {
            value = value.toString();
        }
        if (this.params.format) {
            if (this.params.format === 'date-time') {
                value = new Date(value);
                if (isNaN(value.getDay())) {
                    throw new ValidationError('is not a valid date time');
                }
            }
            else if (this.params.format === 'date') {
                if (!value.match(/^[0-9]{4}-[0-9]{2}-[0-9]{2}$/)) {
                    throw new ValidationError('is not a valid date');
                }
                value = new Date(value);
                if (isNaN(value.getDay())) {
                    throw new ValidationError('is not a valid date');
                }
            }
            // TODO: validate remaining formats
        }
        return value;
    }
    getValue() {
        return {
            ...super.getValue(),
            format: this.params.format,
        };
    }
}
class DateTime extends Str {
    type = 'string';
    constructor(params) {
        super({
            example: '2022-09-15T00:00:00Z',
            ...params,
            format: 'date-time',
        });
    }
}
class Regex extends Str {
    type = 'string';
    constructor(params) {
        super(params);
    }
    validate(value) {
        value = super.validate(value);
        if (!value.match(this.params.pattern)) {
            if (this.params.patternError) {
                throw new ValidationError(`is not a valid ${this.params.patternError}`);
            }
            throw new ValidationError(`does not match the pattern ${this.params.format}`);
        }
        return value;
    }
    getValue() {
        return {
            ...super.getValue(),
            pattern: this.params.pattern,
        };
    }
}
class Email extends Regex {
    type = 'string';
    constructor(params) {
        super({
            pattern: '^[^\\s@]+@[^\\s@]+\\.[^\\s@]+$',
            patternError: 'email',
            ...params,
            format: 'email',
        });
    }
}
class Uuid extends Regex {
    type = 'string';
    constructor(params) {
        super({
            pattern: '^[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}$',
            patternError: 'uuid',
            ...params,
            format: 'uuid',
        });
    }
}
class Hostname extends Regex {
    type = 'string';
    constructor(params) {
        super({
            pattern: '^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\\-]*[a-zA-Z0-9])\\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\\-]*[A-Za-z0-9])$',
            patternError: 'hostname',
            ...params,
            format: 'hostname',
        });
    }
}
class Ipv4 extends Regex {
    type = 'string';
    constructor(params) {
        super({
            pattern: '^(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}$',
            patternError: 'ipv4',
            ...params,
            format: 'ipv4',
        });
    }
}
class Ipv6 extends Regex {
    type = 'string';
    constructor(params) {
        super({
            pattern: '^(?:(?:[a-fA-F\\d]{1,4}:){7}(?:[a-fA-F\\d]{1,4}|:)|(?:[a-fA-F\\d]{1,4}:){6}(?:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|:[a-fA-F\\d]{1,4}|:)|(?:[a-fA-F\\d]{1,4}:){5}(?::(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,2}|:)|(?:[a-fA-F\\d]{1,4}:){4}(?:(?::[a-fA-F\\d]{1,4}){0,1}:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,3}|:)|(?:[a-fA-F\\d]{1,4}:){3}(?:(?::[a-fA-F\\d]{1,4}){0,2}:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,4}|:)|(?:[a-fA-F\\d]{1,4}:){2}(?:(?::[a-fA-F\\d]{1,4}){0,3}:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,5}|:)|(?:[a-fA-F\\d]{1,4}:){1}(?:(?::[a-fA-F\\d]{1,4}){0,4}:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,6}|:)|(?::(?:(?::[a-fA-F\\d]{1,4}){0,5}:(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)(?:\\\\.(?:25[0-5]|2[0-4]\\d|1\\d\\d|[1-9]\\d|\\d)){3}|(?::[a-fA-F\\d]{1,4}){1,7}|:)))(?:%[0-9a-zA-Z]{1,})?$',
            patternError: 'ipv6',
            ...params,
            format: 'ipv6',
        });
    }
}
class DateOnly extends Str {
    type = 'string';
    constructor(params) {
        super({
            example: '2022-09-15',
            ...params,
            format: 'date',
        });
    }
}
class Bool extends Str {
    type = 'boolean';
    validValues = ['true', 'false'];
    validate(value) {
        value = super.validate(value);
        value = value.toLowerCase();
        if (!this.validValues.includes(value)) {
            throw new ValidationError('is not a valid boolean, allowed values are true or false');
        }
        value = value === 'true';
        return value;
    }
}
class Enumeration extends Str {
    isEnum = true;
    values;
    keys;
    constructor(params) {
        super(params);
        let { values } = params;
        if (Array.isArray(values))
            values = Object.fromEntries(values.map((x) => [x, x]));
        this.keys = Object.keys(values);
        this.values = values;
    }
    validate(value) {
        value = super.validate(value);
        if (this.params.enumCaseSensitive !== false) {
            value = this.params.values[value];
        }
        else {
            const key = this.keys.find((key) => key.toLowerCase() === value.toLowerCase());
            value = this.params.values[key];
        }
        if (value === undefined) {
            if (this.params.required === true) {
                throw new ValidationError('is not one of available options');
            }
            // Parameter not required neither one of the available options
            return null;
        }
        return value;
    }
    getValue() {
        return {
            ...super.getValue(),
            enum: this.keys,
        };
    }
}
class Parameter {
    location;
    rawType;
    type;
    params;
    constructor(location, rawType, params) {
        this.location = location;
        this.rawType = rawType;
        if (params.required === undefined)
            params.required = true;
        this.params = params;
        this.type = this.getType(rawType, params);
    }
    getType(type, params) {
        if (type.generated === true) {
            return type;
        }
        if (type.isParameter === true) {
            // @ts-ignore
            return new type({ ...params });
        }
        if (type === String) {
            return new Str({ ...params });
        }
        if (typeof type === 'string') {
            return new Str({ example: type });
        }
        if (type === Number) {
            return new Num({ ...params });
        }
        if (typeof type === 'number') {
            return new Num({ example: type });
        }
        if (type === Boolean) {
            return new Bool({ ...params });
        }
        if (typeof type === 'boolean') {
            return new Bool({ example: type });
        }
        if (type === Date) {
            return new DateTime();
        }
        if (Array.isArray(type)) {
            if (type.length === 0) {
                throw new Error('Arr must have a type');
            }
            return new Arr(this.getType(type[0], params), { ...params });
        }
        if (typeof type === 'object') {
            const parsed = {};
            for (const [key, value] of Object.entries(type)) {
                parsed[key] = this.getType(value, {});
            }
            return new Obj(parsed, params);
        }
        throw new Error(`${type} not implemented`);
    }
    getValue() {
        const schema = removeUndefinedFields(this.type.getValue());
        return {
            description: this.params.description,
            required: this.params.required,
            schema: schema,
            name: this.params.name,
            in: this.location,
        };
    }
    validate(value) {
        if (value === undefined || value === null) {
            if (this.params.default !== undefined && this.params.default !== null) {
                value = this.params.default;
            }
            else {
                if (this.params.required) {
                    throw new ValidationError('is required');
                }
                else {
                    return null;
                }
            }
        }
        value = this.type.validate(value);
        return value;
    }
}
class Body extends Parameter {
    paramsBody;
    constructor(rawType, params) {
        // @ts-ignore
        super(null, rawType, {});
        // @ts-ignore
        this.paramsBody = params;
    }
    getValue() {
        const schema = removeUndefinedFields(this.type.getValue());
        const param = {
            description: this.paramsBody?.description,
            content: {},
        };
        param.content[this.paramsBody?.contentType || 'application/json'] = {
            schema: schema,
        };
        return param;
    }
}
class Resp extends Parameter {
    constructor(rawType, params) {
        // @ts-ignore
        super(null, rawType, params);
    }
    // @ts-ignore
    getValue() {
        const value = super.getValue();
        const contentType = this.params?.contentType
            ? this.params?.contentType
            : 'application/json';
        const param = {
            description: this.params.description || 'Successful Response',
            content: {},
        };
        param.content[contentType] = { schema: value.schema };
        return param;
    }
}
function Query(type, params = {}) {
    return new Parameter('query', type, params);
}
function Path(type, params = {}) {
    return new Parameter('path', type, params);
}
function Header(type, params = {}) {
    return new Parameter('header', type, params);
}
function Cookie(type, params = {}) {
    return new Parameter('cookie', type, params);
}
function extractParameter(request, query, name, location) {
    if (location === 'query') {
        return query[name];
    }
    if (location === 'path') {
        // @ts-ignore
        return request.params[name];
    }
    if (location === 'header') {
        // @ts-ignore
        return request.headers.get(name);
    }
    if (location === 'cookie') {
        throw new Error('Cookie parameters not implemented yet');
    }
}
function extractQueryParameters(request) {
    const url = decodeURIComponent(request.url).split('?');
    if (url.length === 1) {
        return {};
    }
    const query = url.slice(1).join('?');
    const params = {};
    for (const param of query.split('&')) {
        const paramSplitted = param.split('=');
        const key = paramSplitted[0];
        const value = paramSplitted[1];
        if (params[key] === undefined) {
            params[key] = value;
        }
        else if (!Array.isArray(params[key])) {
            params[key] = [params[key], value];
        }
        else {
            params[key].push(value);
        }
    }
    return params;
}
function Required(param) {
    param.params.required = true;
    return param;
}
function removeUndefinedFields(obj) {
    for (const [key, value] of Object.entries(obj)) {
        if (typeof value === 'object' && !Array.isArray(value))
            obj[key] = removeUndefinedFields(value);
        if (value === undefined) {
            delete obj[key];
        }
    }
    return obj;
}
function getFormatedParameters(params) {
    const formated = [];
    const isArray = Array.isArray(params);
    for (const [key, parameter] of Object.entries(params || {})) {
        if (isArray && !parameter.params.name) {
            throw new Error('Parameter must have a defined name when using as Array');
        }
        const name = parameter.params.name ? parameter.params.name : key;
        formated.push({
            // TODO: check this type before assign
            // @ts-ignore
            ...parameter.getValue(),
            name: name,
        });
    }
    return formated;
}

export { Arr, BaseParameter, Body, Bool, Cookie, DateOnly, DateTime, Email, Enumeration, Header, Hostname, Int, Ipv4, Ipv6, Num, Obj, Parameter, Path, Query, Regex, Required, Resp, Str, Uuid, extractParameter, extractQueryParameters, getFormatedParameters, removeUndefinedFields };
