class ApiException extends Error {
    // @ts-ignore
    key;
    constructor(message) {
        super(message);
        this.message = message;
    }
}
class ValidationError extends ApiException {
    constructor(message) {
        super(message);
    }
}

class BaseParameter {
    static isParameter = true;
    isParameter = true;
    type = 'string';
    params;
    generated;
    constructor(params) {
        this.params = params || {};
        this.generated = true;
        if (this.params.required === undefined)
            this.params.required = true;
    }
    getValue() {
        const value = {
            type: this.type,
            description: this.params.description,
            example: this.params.example,
            default: this.params.default,
        };
        if (this.params.deprecated)
            value.deprecated = this.params.deprecated;
        return value;
    }
    validate(value) {
        return value;
    }
}
class Arr extends BaseParameter {
    isArr = true;
    innerType;
    constructor(innerType, params) {
        super(params);
        this.innerType = innerType;
    }
    validate(value) {
        value = super.validate(value);
        if (this.params.required === false && (value === null || value === '')) {
            return null;
        }
        if (Array.isArray(value)) {
            value = value.map((val) => {
                return this.innerType.validate(val);
            });
        }
        else {
            value = [this.innerType.validate(value)];
        }
        return value;
    }
    // @ts-ignore
    getValue() {
        return {
            type: 'array',
            items: this.innerType.getValue(),
        };
    }
}
class Obj extends BaseParameter {
    isObj = true;
    fields;
    constructor(fields, params) {
        super(params); // TODO: fix obj params
        this.fields = fields;
    }
    validate(value) {
        value = super.validate(value);
        for (const [key, param] of Object.entries(this.fields)) {
            try {
                if (value[key] === undefined || value[key] === null) {
                    if (param.params.required) {
                        throw new ValidationError('is required');
                    }
                }
                else {
                    value[key] = param.validate(value[key]);
                }
            }
            catch (e) {
                // @ts-ignore
                e.key = (e.key || '') + `.${key}`;
                throw e;
            }
        }
        return value;
    }
    // @ts-ignore
    getValue() {
        const result = {
            type: 'object',
            properties: {},
        };
        const required = [];
        for (const [key, value] of Object.entries(this.fields)) {
            if (value.params?.required === true) {
                required.push(key);
            }
            if (value.getValue) {
                result.properties[key] = value.getValue();
            }
            else {
                result.properties[key] = value;
            }
        }
        if (required.length > 0) {
            result.required = required;
        }
        if (this.params.xml) {
            result.xml = this.params.xml;
        }
        return result;
    }
}
class Num extends BaseParameter {
    type = 'number';
    validate(value) {
        value = super.validate(value);
        value = Number.parseFloat(value);
        if (isNaN(value)) {
            throw new ValidationError('is not a valid number');
        }
        return value;
    }
}
class Str extends BaseParameter {
    type = 'string';
    constructor(params) {
        super(params);
    }
    validate(value) {
        value = super.validate(value);
        if (typeof value !== 'string') {
            value = value.toString();
        }
        if (this.params.format) {
            if (this.params.format === 'date-time') {
                value = new Date(value);
                if (isNaN(value.getDay())) {
                    throw new ValidationError('is not a valid date time');
                }
            }
            else if (this.params.format === 'date') {
                if (!value.match(/^[0-9]{4}-[0-9]{2}-[0-9]{2}$/)) {
                    throw new ValidationError('is not a valid date');
                }
                value = new Date(value);
                if (isNaN(value.getDay())) {
                    throw new ValidationError('is not a valid date');
                }
            }
            // TODO: validate remaining formats
        }
        return value;
    }
    getValue() {
        return {
            ...super.getValue(),
            format: this.params.format,
        };
    }
}
class DateTime extends Str {
    type = 'string';
    constructor(params) {
        super({
            example: '2022-09-15T00:00:00Z',
            ...params,
            format: 'date-time',
        });
    }
}
class Bool extends Str {
    type = 'boolean';
    validValues = ['true', 'false'];
    validate(value) {
        value = super.validate(value);
        value = value.toLowerCase();
        if (!this.validValues.includes(value)) {
            throw new ValidationError('is not a valid boolean, allowed values are true or false');
        }
        value = value === 'true';
        return value;
    }
}
class Parameter {
    location;
    rawType;
    type;
    params;
    constructor(location, rawType, params) {
        this.location = location;
        this.rawType = rawType;
        if (params.required === undefined)
            params.required = true;
        this.params = params;
        this.type = this.getType(rawType, params);
    }
    getType(type, params) {
        if (type.generated === true) {
            return type;
        }
        if (type.isParameter === true) {
            // @ts-ignore
            return new type({ ...params });
        }
        if (type === String) {
            return new Str({ ...params });
        }
        if (typeof type === 'string') {
            return new Str({ example: type });
        }
        if (type === Number) {
            return new Num({ ...params });
        }
        if (typeof type === 'number') {
            return new Num({ example: type });
        }
        if (type === Boolean) {
            return new Bool({ ...params });
        }
        if (typeof type === 'boolean') {
            return new Bool({ example: type });
        }
        if (type === Date) {
            return new DateTime();
        }
        if (Array.isArray(type)) {
            if (type.length === 0) {
                throw new Error('Arr must have a type');
            }
            return new Arr(this.getType(type[0], params), { ...params });
        }
        if (typeof type === 'object') {
            const parsed = {};
            for (const [key, value] of Object.entries(type)) {
                parsed[key] = this.getType(value, {});
            }
            return new Obj(parsed, params);
        }
        throw new Error(`${type} not implemented`);
    }
    getValue() {
        const schema = removeUndefinedFields(this.type.getValue());
        return {
            description: this.params.description,
            required: this.params.required,
            schema: schema,
            name: this.params.name,
            in: this.location,
        };
    }
    validate(value) {
        if (value === undefined || value === null) {
            if (this.params.default !== undefined && this.params.default !== null) {
                value = this.params.default;
            }
            else {
                if (this.params.required) {
                    throw new ValidationError('is required');
                }
                else {
                    return null;
                }
            }
        }
        value = this.type.validate(value);
        return value;
    }
}
class Body extends Parameter {
    paramsBody;
    constructor(rawType, params) {
        // @ts-ignore
        super(null, rawType, {});
        // @ts-ignore
        this.paramsBody = params;
    }
    getValue() {
        const schema = removeUndefinedFields(this.type.getValue());
        const param = {
            description: this.paramsBody?.description,
            content: {},
        };
        param.content[this.paramsBody?.contentType || 'application/json'] = {
            schema: schema,
        };
        return param;
    }
}
class Resp extends Parameter {
    constructor(rawType, params) {
        // @ts-ignore
        super(null, rawType, params);
    }
    // @ts-ignore
    getValue() {
        const value = super.getValue();
        const contentType = this.params?.contentType
            ? this.params?.contentType
            : 'application/json';
        const param = {
            description: this.params.description || 'Successful Response',
            content: {},
        };
        param.content[contentType] = { schema: value.schema };
        return param;
    }
}
function extractParameter(request, query, name, location) {
    if (location === 'query') {
        return query[name];
    }
    if (location === 'path') {
        // @ts-ignore
        return request.params[name];
    }
    if (location === 'header') {
        // @ts-ignore
        return request.headers.get(name);
    }
    if (location === 'cookie') {
        throw new Error('Cookie parameters not implemented yet');
    }
}
function extractQueryParameters(request) {
    const url = decodeURIComponent(request.url).split('?');
    if (url.length === 1) {
        return {};
    }
    const query = url.slice(1).join('?');
    const params = {};
    for (const param of query.split('&')) {
        const paramSplitted = param.split('=');
        const key = paramSplitted[0];
        const value = paramSplitted[1];
        if (params[key] === undefined) {
            params[key] = value;
        }
        else if (!Array.isArray(params[key])) {
            params[key] = [params[key], value];
        }
        else {
            params[key].push(value);
        }
    }
    return params;
}
function removeUndefinedFields(obj) {
    for (const [key, value] of Object.entries(obj)) {
        if (typeof value === 'object' && !Array.isArray(value))
            obj[key] = removeUndefinedFields(value);
        if (value === undefined) {
            delete obj[key];
        }
    }
    return obj;
}
function getFormatedParameters(params) {
    const formated = [];
    const isArray = Array.isArray(params);
    for (const [key, parameter] of Object.entries(params || {})) {
        if (isArray && !parameter.params.name) {
            throw new Error('Parameter must have a defined name when using as Array');
        }
        const name = parameter.params.name ? parameter.params.name : key;
        formated.push({
            // TODO: check this type before assign
            // @ts-ignore
            ...parameter.getValue(),
            name: name,
        });
    }
    return formated;
}

class OpenAPIRoute {
    static isRoute = true;
    static schema;
    params;
    constructor(params) {
        this.params = params;
    }
    static getSchema() {
        return this.schema;
    }
    get schema() {
        // @ts-ignore
        return this.__proto__.constructor.schema;
    }
    getSchema() {
        // @ts-ignore
        return this.__proto__.constructor.getSchema();
    }
    static getParsedSchema() {
        const schema = this.getSchema();
        let requestBody = null;
        if (schema.requestBody) {
            requestBody = new Body(schema.requestBody, {
                contentType: schema.requestBody.contentType,
            }).getValue();
        }
        const responses = {};
        if (schema.responses) {
            for (const [key, value] of Object.entries(schema.responses)) {
                const resp = new Resp(value.schema, value);
                responses[key] = resp.getValue();
            }
        }
        // Deep copy
        return {
            ...schema,
            parameters: schema.parameters
                ? getFormatedParameters(schema.parameters)
                : [],
            responses: responses,
            ...(requestBody ? { requestBody: requestBody } : {}),
        };
    }
    jsonResp(params) {
        return new Response(JSON.stringify(params.data), {
            headers: {
                'content-type': 'application/json;charset=UTF-8',
            },
            status: params.status || 200,
        });
    }
    handleValidationError(errors) {
        return this.jsonResp({
            data: {
                errors: errors,
                success: false,
                result: {},
            },
            status: 400,
        });
    }
    async execute(...args) {
        const { data, errors } = await this.validateRequest(args[0]);
        if (Object.keys(errors).length > 0) {
            return this.handleValidationError(errors);
        }
        args.push(data);
        const resp = await this.handle(...args);
        if (!(resp instanceof Response) && typeof resp === 'object') {
            return this.jsonResp({ data: resp });
        }
        return resp;
    }
    async validateRequest(request) {
        const params = this.getSchema().parameters || {};
        const requestBody = this.getSchema().requestBody;
        const queryParams = extractQueryParameters(request);
        const validatedObj = {};
        const validationErrors = {};
        // check for query unknown parameters
        if (this.params?.raiseUnknownParameters) {
            for (const key of Object.keys(queryParams)) {
                if (Array.isArray(params)) {
                    let exists = false;
                    for (const param of params) {
                        if (param.params.name === key) {
                            exists = true;
                            break;
                        }
                    }
                    if (!exists) {
                        validationErrors[key] = `is an unknown parameter`;
                    }
                }
                else {
                    if (params[key] === undefined) {
                        validationErrors[key] = `is an unknown parameter`;
                    }
                }
            }
        }
        for (const [key, value] of Object.entries(params)) {
            // @ts-ignore
            const param = value;
            const name = param.params.name ? param.params.name : key;
            const rawData = extractParameter(request, queryParams, name, param.location);
            try {
                validatedObj[name] = param.validate(rawData);
            }
            catch (e) {
                validationErrors[name] = e.message;
            }
        }
        if (request.method.toLowerCase() !== 'get' && requestBody) {
            let json;
            let loaded = false;
            try {
                // @ts-ignore
                json = await request.json();
                loaded = true;
            }
            catch (e) {
                validationErrors['body'] = e.message;
            }
            if (loaded)
                try {
                    validatedObj['body'] = new Body(requestBody).validate(json);
                }
                catch (e) {
                    validationErrors['body' + e.key] = e.message;
                }
        }
        return {
            data: validatedObj,
            errors: validationErrors,
        };
    }
    handle(...args) {
        throw new Error('Method not implemented.');
    }
}

export { OpenAPIRoute };
